# -*- coding: utf-8 -*-
"""
Created on Tue Jun  8 09:35:46 2021

@author: sparrenberg
"""
import numpy as np
from scipy import sparse
from numba import njit, float64, int64
from scipy.optimize import least_squares
import pickle

# SMD class which deals with fluorescence fluctuation measurement data
# Object attributes are specific for FluoroDrop FD1 setup and obtained by calibration measurements
# Object attributes for simulations are specific for simulation conditions used in dissertation Sparrenberg (2021)

class SMD():
    def __init__(self, mod = 'meas', **kwargs):
        if mod == 'meas':
            # Standard attribute values for measurement instance
            self.name =  kwargs.get('name', None)
            self.date = kwargs.get('date', None)
            self.device = kwargs.get('device', 'FluoroDrop 1')
            self.device_const = kwargs.get('r', 19.8)# z0/r0
            self.wxy = kwargs.get('wxy', 0.40) # µm
            self.exc_power = kwargs.get('power', 2.5) # µW 
            self.wavelength = kwargs.get('wavelength', 488) # nm
            self.objective = kwargs.get('objective', '63x/0.75 LD Neofluar')
            self.filter = kwargs.get('filter', '535/50') 
            self.mtp = kwargs.get('mtp', 'Greiner BioOne, uCLear, Non-Binding') 
            self.well = kwargs.get('well', None)
            self.meas_num = kwargs.get('meas_num', None)
            self.conc = kwargs.get('conc', None)
            self.soluble = kwargs.get('soluble', None)
            self.bg = kwargs.get('bg', 0) # kCps
            self.sample = kwargs.get('sample', 'dye')
            self.temperatur = kwargs.get('temp', 298) # K
        else:
            # Standard attribute values for simulation instance
            self.name =  kwargs.get('name', None)
            self.device = kwargs.get('device', 'Monte-Carlo simulation')
            self.version = kwargs.get('version', 'MCSim_v1p13.exe')
            self.sim_steps = kwargs.get('steps', None)
            self.mt = kwargs.get('mt', None)
            self.wxy = kwargs.get('wxy', 0.40) # µm
            self.wz = kwargs.get('wz', 0.40) # µm
            self.device_const = kwargs.get('r', 1)
            self.lattice_x = kwargs.get('lattice_x', 4.8) # µm
            self.lattice_y = kwargs.get('lattice_y', 4.8) # µm
            self.lattice_z = kwargs.get('lattice_z', 4.8) # µm
            self.transmission = kwargs.get('trans', 1.0)
            self.meas_num = kwargs.get('meas_num', None)
            self.bg = kwargs.get('bg', 0) # kCps
            self.deadtime = kwargs.get('dt', 0) # ns
            self.afterpulsing_prob = kwargs.get('ap_prob', 0)
            self.afterpulsing_decay = kwargs.get('ap_decay', 0) # s
            self.sample = kwargs.get('sample', 'MCS')
            self.num_molecules = kwargs.get('molecules', None)
            self.diffusion_coef = kwargs.get('D', None) # µm^2/s
            self.light_factor = kwargs.get('LF', None)
            self.saturation_rate = kwargs.get('sat', None)
            self.bleach_lim = kwargs.get('bleach_lim', None)
            
    def loadfile(self, file, mt = 10):
        '''Importing the measurement data from compressed npz file'''
        
        # deals automatically with file endings 'dat' and 'npz'
        
        # file type from simulation
        if file[-3:] == 'dat': 
            data = np.fromfile(file, dtype = int)
            self.counts = sparse.csr_matrix(data)
            self.mt = mt
            
        # file type from measurement or converted dat file
        elif file[-3:] == 'npz':    
            s_data = sparse.load_npz(file)
            self.counts = s_data[0,:-1]
            self.mt = s_data[0,-1]
        else:
            print('File type not supported')
            
        # mean intensity in kCps
        self.mean = round(np.mean(self.counts) * self.counts.shape[1]/(1e3 * self.mt), 2)
        
    def get_counts(self):
        '''Get raw fluorescence fluctuation counts of measurement'''
        data = self.counts.toarray().flatten()
        return data
    
    def FCS(self, binning = 1e-6, num_bins = 95, shifter = 2):
        '''Calculates autocorrelation of FCS on data'''
        
        # get raw counts
        x = self.get_counts()
        
        # entries per second and mean
        n_values = int(len(x) / self.mt) 
        mean = np.mean(x)
        
        #Calculates autocorrelation
        corr = np.fft.ifft(np.abs(np.fft.fft(x))**2).real
        corr_norm = corr[:n_values] / (len(x) * mean**2) - 1 #returns data for 1 sec
        
        # filters data by median to get nearly log scaled lag times
        time, fcs = SMD.__FCS_filter(corr_norm, binning, num_bins)
        
        # Add FCS time and data to class instance
        self.FCS_time = time[shifter:]
        self.FCS_data = fcs[shifter:]
               
    def __FCS_filter(corr_data, binning, num_bins):
        '''Filtering data using a moving median filter with logarithmic scaling'''
        
        # Set log scale with special handling for short lag times
        log = SMD.__log_space(len(corr_data), num_bins)
        
        # Empty arrays for FCS data and lag time
        medfilt = np.zeros(len(log))
        medfilt_time = np.zeros(len(log))
        
        # Span lag times aligning to input corr_data
        time = np.linspace(binning, 1, len(corr_data), endpoint = True)
        
        # Loop through log scaled space
        for i in range(len(log)):
            # no difference for first entry
            if i == 0:
                diff = 0
                
            # difference for all other entries
            else: 
                diff = int(log[i]) - int(log[i - 1]) - 1
                
            # filtered data is corr_data for all zero differences
            if diff == 0:
                medfilt[i] = corr_data[int(log[i])]
                medfilt_time[i] = time[int(log[i])]
                
            # otherwise get median for data spanned by difference while length of 
            # corr_data is not exceeded
            elif int(log[i]) + diff < len(corr_data):    
                medfilt[i] = np.median(corr_data[(int(log[i]) - diff):(int(log[i]) + diff)])
                medfilt_time[i] = time[int(log[i])]
                
            # deal with data at the end of array and calculate median 
            # of remaining data
            else:
                medfilt[i] = np.median(corr_data[(int(log[i]) - diff): ])
                medfilt_time[i] = time[int(log[i])]
                
        # return filtered lag time and FCS data
        return medfilt_time, medfilt 
    
    def __log_space(limit, n):
        '''Subdivide data in logarithmic spaced subsamples
        but keeping short lag times as they are'''
        
        # start of results list
        result = [1]
        
        # check to avoid ZeroDivisionError
        if n > 1:  
            ratio = (float(limit) / result[-1]) ** (1.0 / (n - len(result)))
        
        # continue until results list is equal to desired number of entries
        while len(result) < n:
            next_value = result[-1] * ratio
            if next_value - result[-1] >= 1:
                result.append(next_value) # next_value will be a different integer
            else:
                # problem! same integer. Incrementing previous value
                result.append(result[-1] + 1)
                
                # recalculate the ratio so that the remaining values 
                # will scale correctly
                ratio = (float(limit) / result[-1]) ** (1.0 / (n - len(result)))
                
        # round, re-adjust to 0 indexing (i.e. minus 1) and return np.uint64 array
        return np.array(list(map(lambda x: round(x) -1, result)), dtype = np.uint64)
    
    def mSMR(self, p_ap = 0.006, d_time = 70e-9, shifter = 2):
        '''Calculates autocorrelation of FCS on data.
        Afterpulsing probability and dead time are set to default values from
        data sheet of photodiode'''
        
        # get raw counts
        x = self.get_counts()
        
        # Get measurement time
        mt = self.mt
        
        # find all divisors for data to ensure equal bin size
        divisors = SMD.__divisor_generator(len(x))
        
        # Set up arrays for sampling time and mu to fill
        mu = np.zeros(len(divisors)) 
        sample_time = np.zeros(len(divisors))  
    
        # Prior calculation of some binned data to accelerate later binning
        bin2 = SMD.bin_array(x, int(len(x) / 2))
        bin5 = SMD.bin_array(x, int(len(x) / 5))
        bin50 = SMD.bin_array(bin5, int(len(x) / 50))
        bin500 = SMD.bin_array(bin50, int(len(x) / 500))
    
        # calculate mu for binned data
        for idx, div in enumerate(divisors):
            # Get current number of bins and sampling time
            bins_temp = int(len(x) / div)
            sample_time_temp = mt / bins_temp
            
            # Use pre-calculated binned data for fast calculation of mSMR
            if div%500 == 0:
                M = SMD.__moments(SMD.bin_array(bin500, bins_temp))
            elif div%50 == 0:
                M = SMD.__moments(SMD.bin_array(bin50, bins_temp))
            elif div%5 == 0:
                M = SMD.__moments(SMD.bin_array(bin5, bins_temp))            
            elif div%2 == 0:
                M = SMD.__moments(SMD.bin_array(bin2, bins_temp))
            else:
                M = SMD.__moments(SMD.bin_array(x, bins_temp)) 
                
            # Correct first and second moments by afterpulsing and dead time
            # using a first order correction
            delta = d_time/sample_time_temp
            M1 = M[0] * (1 - p_ap - delta) + delta * M[1]
            M2 = M[1] * (1 - 2 * p_ap - 3 * delta) + 2 * delta * M[2] + (delta - p_ap) * M[0]
            
            # Calculate mu(T) from moments
            mu_temp = (M2 - M1**2 - M1)/(M1 * sample_time_temp )
            
            # add current mu(T) and sample time (T) to arrays
            mu[idx] = mu_temp
            sample_time[idx] = sample_time_temp
        self.mSMR_time = sample_time[shifter:]
        
        # Add mSMR time and data to class instance
        self.mSMR_data = mu[shifter:] / 1000 # in kCps
        self.artefacts = [p_ap, d_time]
            
    @njit(float64[:](int64)) 
    def __divisor_generator(n):
        '''gives all even divisors of a number'''
        
        small_divisors = []
        large_divisors = []
        for i in range(1, int(n**0.5 + 1)):
            if n % i == 0:
                small_divisors.append(i)
                if i * i != n:
                    large_divisors.append(int(n / i))
        return np.array(small_divisors + large_divisors[1:][::-1]).astype(np.float64)

    @njit(float64[:](float64[:]))
    def __moments(data):
        '''Gives the first three moments of an array'''
        
        M1 = 0
        M2 = 0
        M3 = 0
        for i in range(len(data)):
            M1 += data[i]
            M2 += data[i]**2
            M3 += data[i]**3
        return np.array([M1 / len(data), M2 / len(data), M3 / len(data)])
    
    def bin_array(data, bins):
        '''Bins an array to given number of bins.
        Bins need to be even divisors of len(data).
        Fast binning by using matrix multiplication'''
        
        if len(data) == bins:
            return data.astype(np.float64)
        else:
            width = len(data) // bins
            
            # Bin by using matrix multiplication
            return (data.reshape(bins, width) @ np.ones((width, 1))).reshape(bins)
        
    def FCS_mod(x, t, r):
        '''Autocorrelation function for non linear fit of FCS data'''
        G_diff = (1 / (x[0])) * (1 / (1 + t / x[1])) * (1 / (1 + (r**(-2)) * t / x[1])**0.5)
        
        # Case one: triplet parameters are given
        if len(x) == 4:
            G_triplet = (1 + (x[2] / (1 - x[2])) * np.exp(-t / x[3]))
            return G_diff * G_triplet
        
        # Case two: only parameters of diffusion model are given
        elif len(x) == 2:
            return G_diff
    
    def __B2_mod(tau,t, r):
        '''Definition of B2 according to Palo et al. (2000) from Scales (2019) 
        where µ0 ist the single molecule brightness'''
        
        # Helper parameters
        a = t / tau
        b = 1 / r**2
        
        # Model for symmetric case of r0 == z0
        if r == 1:
            B2 = 8 / (a**2) * (a / 2 - (a + 1)**0.5 + 1)
            return B2
        
        # Model for asymmetric case
        elif r > 1:    
            B2 = (4 /(a**2 * b)) * ((b * (1 + a) / (np.sqrt(1 - b))) * np.arctanh((np.sqrt(1 - b)) * ((np.sqrt(1 + b * a) -1)/(b + np.sqrt(1 + b * a) -1))) + 1-np.sqrt(1 + b * a))    
            return B2
        
        # r < 1 is not defined
        else:
            return -1
        
    def __triplet_mod(t, F, tF):
        '''Triplet model and simple on/off isomerization model for
        mSMR analysis'''
        
        # Deals with case tF = 0 which causes ZeroDivisionError and returns 
        # neutral element
        if tF == 0:
            triplet = np.ones(len(t))
        
        # Triplet model
        else: 
            triplet = 1 + 2/t**2 * (F * t * tF + (-1 + np.exp(-t / tF))* F * tF**2)/(1-F)
        return triplet
  
    def mSMR_mod(x, t, r):
        '''mSMR model according to Palo et al. (2000) from Scales (2019)'''
        
        # divide mu0 by 1000 to achieve better fitting results
        mu0 = x[0] * 1000 
        
        # Gamma parameter from Gaussian ellipsoid
        gamma = 1 / (2 * np.sqrt(2))
        
        # Case one: triplet parameters are given
        if len(x) == 4:
            Q = gamma * mu0 * SMD.__B2_mod(x[1], t, r) * SMD.__triplet_mod(t, x[2], x[3])
        # Case two: only parameters of diffusion model are given
        elif len(x) == 2:
            Q = gamma * mu0 * SMD.__B2_mod(x[1], t, r)
        return Q
   
    def Error_func(x, t, y, r, mod = 'FCS'):
        '''Error function of FCS and mSMR model for data fitting'''
        
        if mod == 'FCS':
            return SMD.FCS_mod(x, t, r) - y
        elif mod == 'mSMR':
            return SMD.mSMR_mod(x, t, r) - y
        
    def fit_data(self, mod = 'FCS', on_corr = False):
        '''Data fitting of FCS and mSMR data.
        Three mods are given:
        "dye": fitting of dyes such as Alexa 488
        "DNA": fitting of slow and bright polymers such as DNA
        "MCS": fitting of Monte-Carlo simulated data'''
        
        # Fitting parameters for mSMR and FCS data
        # Parameters are vary for differnt experimental setups
        if self.sample == 'dye':
            x = [23.0, 0.0002, 0.1, 4e-6]
            x_l = [0.0, 0.0, 0, 1e-6]
            x_h = [np.inf, 0.0003, 0.4, 2e-5]
        elif self.sample == 'DNA':
            x = [0.1, 0.001, 0.1, 2e-5]
            x_l = [0.0, 0.0, 0, 1e-5]
            x_h = [np.inf, 0.02, 0.5, 1e-4]
        elif self.sample == 'MCS':
            x = [23.0, 0.0002]
            x_l = [0.0, 0.0]
            x_h = [np.inf, 0.003]
        
        # Fitting FCS data
        if mod == 'FCS':
            time = self.FCS_time
            
            # Fitting raw FCS data
            if on_corr == False:
                data = self.FCS_data
                fit = least_squares(SMD.Error_func, x, loss = 'soft_l1', 
                                    f_scale = 0.1,
                                    args = (time, data, self.device_const, mod), 
                                    bounds = (x_l, x_h))
                self.FCS_x = fit.x
                self.N_FCS = fit.x[0]
                self.Tau_FCS = fit.x[1]
                
                # Try to add triplet parameters
                try:
                    self.F_FCS = fit.x[2]
                    self.TF_FCS = fit.x[3]
                except:
                    self.F_FCS = None
                    self.TF_FCS = None
            
            # Fitting background corrected FCS data
            else:
                data = self.FCS_data_corr
                fit = least_squares(SMD.Error_func, x, loss = 'soft_l1', 
                                    f_scale = 0.1,
                                    args = (time, data, self.device_const, mod), 
                                    bounds = (x_l, x_h))
                self.FCS_c_x = fit.x 
                self.N_c_FCS = fit.x[0]
                self.Tau_c_FCS = fit.x[1]
                
                # Try to add triplet parameters
                try:
                    self.F_c_FCS = fit.x[2]
                    self.TF_c_FCS = fit.x[3]
                except:
                    self.F_c_FCS = None
                    self.TF_c_FCS = None
                    
        # Fitting mSMR data
        elif mod == 'mSMR':
            time = self.mSMR_time
            
            # Fitting raw mSMR data
            if on_corr == False:
                data = self.mSMR_data
                fit = least_squares(SMD.Error_func, x, loss = 'soft_l1', 
                                    f_scale = 0.1,
                                    args = (time, data, self.device_const, mod),
                                    bounds = (x_l, x_h))
                self.mSMR_x = fit.x 
                self.mu0_mSMR = fit.x[0]
                self.Tau_mSMR = fit.x[1]
                
                # Try to add triplet parameters
                try:
                    self.F_mSMR = fit.x[2]
                    self.TF_mSMR = fit.x[3]
                except:
                    self.F_mSMR = None
                    self.TF_mSMR = None
                    
            # Fitting background corrected mSMR data
            else:
                data = self.mSMR_data_corr
                fit = least_squares(SMD.Error_func, x, loss = 'soft_l1', 
                                    f_scale = 0.1,
                                    args = (time,data, self.device_const, mod), 
                                    bounds = (x_l, x_h))
                self.mSMR_c_x = fit.x
                self.mu0_c_mSMR = fit.x[0]
                self.Tau_c_mSMR = fit.x[1]
                
                # Try to add triplet parameters
                try:
                    self.F_c_mSMR = fit.x[2]
                    self.TF_c_mSMR = fit.x[3]
                except:
                    self.F_c_mSMR = None
                    self.TF_c_mSMR = None
        else:
            print('mod not supported, try mod = "FCS" or "mSMR"')
        
    def bg_correction(self, mod = 'FCS', on_val = True):
        '''Apply background correction to FCS or mSMR data.
        on_val = True corrects the fitting parameters.
        on_val = False corrects the FCS or mSMR data directly'''
        
        # Get background and mean of measurement
        bg = self.bg
        mean = self.mean
        
        # Deals with ZeroDivisionError
        if bg == mean:
            print('mean counts equal background noise. No correction applied')
        
        # Correct data
        else: 
            # Correct FCS data
            if mod == 'FCS':
                chi_FCS = (1 + bg / (mean - bg))**2
                
                # Case one: correct FCS data
                if on_val == False:
                    self.FCS_data_corr = self.FCS_data * chi_FCS
                    
                # Case two: correct FCS parameters from fit
                else:
                    self.N_corr = self.FCS_x[0] / chi_FCS 
                    
            # Correct FCS data
            elif mod == 'mSMR':
                chi_mSMR = (1 - bg / mean)**-1
                
                # Case one: correct mSMR data
                if on_val == False:
                    self.mSMR_data_corr = self.mSMR_data * chi_mSMR
                    
                # Case two: correct mSMR parameters from fit
                else:
                    self.mu0_corr = self.mSMR_x[0] * chi_mSMR
            else:
                print('mod not supported, try mod = "FCS" or "mSMR"')
        
    def save(self, file):
        '''Saves instance to pickle file'''
        with open('{}.pkl'.format(file), 'wb') as output:
            pickle.dump(self, output, pickle.HIGHEST_PROTOCOL)
            
    def load(file):
        '''Loads pickle file'''
        with open(file, 'rb') as input:
            return pickle.load(input)


import glob
import pandas as pd

# Class to ease analysis of fluorescence fluctuation measurements
# Needs pickled measurement files

class Experiment():
    def __init__(self, path, ending = '.pkl'):
        '''Get all pickle files in directory during initialization'''
        self.directory = path    
        self.files = glob.glob('{}\*{}'.format(self.directory, ending))
        
    def sum_results(self, *args): 
        '''Accepts unzipped list of values to gather.
        E.g. self.sum_results(*['mean','mu0','N_corr'])'''
        
        # Init empty array for data
        tmp_data = np.empty((len(self.files), len(args)), float)
        
        # loop through files in instance
        for idx,f in enumerate(self.files):
            # Load file
            exp = SMD.load(f)
            
            # Get requested values from loaded intance
            row = [vars(exp)[arg] for arg in args]
            
            # Add data to array
            tmp_data[idx] = row
        
        # Put everything in a dataframe
        data = pd.DataFrame(tmp_data) 
        
        # Rename columns
        data.columns = list(args)
        
        # Return dataframe
        return data
    
    def sum_series(self, mod = 'FCS'):
        '''Returns FCS or mSMR data instead of fitting results 
        as self.sum_results() does.
        Can return background noise corrected or uncorrected data 
        of FCS or mSMR analysis.'''
        
        # setup instance 
        # mods contains accepted mod parameters in function call
        # instances contains attribute names as used in pickle file
        # series is correct attribute name according to function call
        mods = ['FCS', 'FCS_corr', 'mSMR', 'mSMR_corr']
        instances = ['FCS_data', 'FCS_data_corr', 'mSMR_data', 'mSMR_data_corr']
        series = instances[mods.index(mod)]
        
        # Setup lists for series to be summarized
        conc = np.empty(len(self.files), float)
        power = np.empty(len(self.files), float)
        
        # Loop through files
        for idx, f in enumerate(self.files):
            
            # Load file
            exp = SMD.load(f)
            
            # Get lagor sampling time and sett up empty array for mSMR or FCS data
            # Assumes all data to be of equal length
            if idx == 0:
                time = exp.FCS_time if mod in ['FCS', 'FCS_corr'] else exp.mSMR_time
                tmp_data = np.empty((len(self.files), len(time)), float)
            
            # Add FCS or mSMR data
            tmp_data[idx] = vars(exp)[series]
            
            # Add concentration and excitation power
            conc[idx] = exp.conc
            power[idx] = exp.exc_power
            
        # Convert array to dataframe, add column names and power and concentration
        df = pd.DataFrame(tmp_data)    
        df.columns = time
        df['exc_power'] = power
        df['conc'] = conc
        
        # return data frame
        return df
    
    def sum_MCS(self, mod = 'FCS', sorter = None, num_meas = 5):
        '''Deals with mSMR or FCS data of unequal length as in some
        Monte-Carlo simulations.
        Returns mean of n fold measurement.'''
        
        # setup instance 
        # mods contains accepted mod parameters in function call
        # instances contains attribute names as used in pickle file
        # series is correct attribute name according to function call
        mods = ['FCS', 'FCS_corr', 'mSMR', 'mSMR_corr']
        instances = ['FCS_data','FCS_data_corr','mSMR_data','mSMR_data_corr']
        series = instances[mods.index(mod)]
        
        # Assign correct time string to loaded data
        t = 'FCS_time' if mod in ['FCS', 'FCS_corr'] else 'mSMR_time'
        
        # We need to deal with data of different length
        # Adding them using pd.concat --> fills empty spaces with NaNs
        df = pd.DataFrame()
        dt = pd.DataFrame()
        
        # Loop through files
        for idx, f in enumerate(self.files):
            # load file
            exp = SMD.load(f)
            
            # get time and data of measurement
            time = pd.DataFrame(vars(exp)[t], columns = [idx])
            data = pd.DataFrame(vars(exp)[series], columns = [idx])
            df = pd.concat([df, data], axis = 1)
            dt = pd.concat([dt, time], axis = 1)
            
        # Calculate mean of num_meas measurments
        mean = df.groupby(np.arange(len(df.columns)) // num_meas, axis = 1).mean()
        time = dt.groupby(np.arange(len(dt.columns)) // num_meas, axis = 1).mean()
        
        # return time and mean data of FCS or mSMR analysis
        return time, mean